//==============================================================================
//
// Copyright (c) 2007 Ryan Winter. All rights reserved.
//
// This file may be used under the terms of the GNU General Public
// License version 2.0 appearing in the file LICENSE packaged with this file.
//
//==============================================================================

#include "mandelbrot.h"

// number of pixels in a fractal unit
const double RATIO = 200;

const double MIN_X = -2.5;
const double MAX_X =  1.0;
const double MIN_Y = -1.5;
const double MAX_Y =  1.5;

//------------------------------------------------------------------------------
// public:
//------------------------------------------------------------------------------
Mandelbrot::Mandelbrot()
{
}

//------------------------------------------------------------------------------
Mandelbrot::~Mandelbrot()
{
}

//------------------------------------------------------------------------------
QRect
Mandelbrot::limits(double zoom) const
{
    int min_x = static_cast<int>(MIN_X * RATIO * zoom);
    int max_x = static_cast<int>(MAX_X * RATIO * zoom);
    int min_y = static_cast<int>(MIN_Y * RATIO * zoom);
    int max_y = static_cast<int>(MAX_Y * RATIO * zoom);

    return QRect(QPoint(min_x, min_y), QPoint(max_x, max_y));
}

//------------------------------------------------------------------------------
void
Mandelbrot::generate(const CellDescriptor&      cell,
                     std::vector<unsigned int>& data) const
{
    const double min_re = cell.x * cell.width  / RATIO / cell.zoom;
    const double min_im = cell.y * cell.height / RATIO / cell.zoom;
    const double max_re = min_re + cell.width  / RATIO / cell.zoom;
    const double max_im = min_im + cell.height / RATIO / cell.zoom;

    const double delta_re = (max_re - min_re) / (cell.width  - 1);
    const double delta_im = (max_im - min_im) / (cell.height - 1);

    for (int yy = 0; yy < cell.height; ++yy)
    {
        double c_im = min_im + yy * delta_im;

        for (int xx = 0; xx < cell.width; ++xx)
        {
            double c_re = min_re + xx * delta_re;

            double z_re = c_re;
            double z_im = c_im;

            double z_re2      = 0;
            double z_im2      = 0;
            int    iterations = 0;

            do
            {
                z_re2 = z_re * z_re;
                z_im2 = z_im * z_im;

                z_im = 2 * z_re * z_im + c_im;
                z_re = z_re2 - z_im2 + c_re;

                ++iterations;
            }
            while (z_re2 + z_im2 < 4.0 && iterations < cell.iterations);

            if (cell.iterations == iterations) iterations = 0;

            data[yy * cell.width + xx] = iterations;
        }
    }
}

