//==============================================================================
//
// Copyright (c) 2007 Ryan Winter. All rights reserved.
//
// This file may be used under the terms of the GNU General Public
// License version 2.0 appearing in the file LICENSE packaged with this file.
//
//==============================================================================

#include "fractal_thread_pool.h"

#include <iostream>

//------------------------------------------------------------------------------
// public:
//------------------------------------------------------------------------------
FractalThreadPool::FractalThreadPool()
    :
    m_lock(QMutex::Recursive),
    m_data_access(),
    m_threads(),
    m_cells()
{
    setCpuNumber(1);
}

//------------------------------------------------------------------------------
FractalThreadPool::~FractalThreadPool()
{
    clearThreads();
    m_cells.clear();
}

//------------------------------------------------------------------------------
void
FractalThreadPool::setCpuNumber(int cpus)
{
    QMutexLocker ml(&m_lock);

    if (static_cast<int>(m_threads.size()) == cpus) return;

    clearThreads();

    std::cout << "cpus=" << cpus << std::endl;

    // create all the new threads
    for (int ii = 0; ii < cpus + 1; ++ii)
    {
        std::cout << "cpu=" << ii << std::endl;
        FractalWidgetThread *thread
            = new FractalWidgetThread(m_cells, m_lock, m_data_access);

        // forward the signals
        connect(thread, SIGNAL(threadComplete(CellImage*)),
                this,   SIGNAL(processingComplete(CellImage*)));

        // start the thread at low priority so it doesn't impact the GUI
        thread->start();
        thread->setPriority(QThread::LowestPriority);

        m_threads.push_back(thread);
    }
}

//------------------------------------------------------------------------------
void
FractalThreadPool::queue(const CellDescriptor& cell)
{
    QMutexLocker ml(&m_lock);

    if (!m_cells.empty())
    {
        // check if the version has changed
        if (m_cells.begin()->version != cell.version)
        {
            m_cells.clear();

            // reset the semaphore
            m_data_access.acquire(m_data_access.available());
        }
    }

    m_cells.push_back(cell);

    m_data_access.release();
}

//------------------------------------------------------------------------------
// private:
//------------------------------------------------------------------------------
void
FractalThreadPool::clearThreads()
{
    // wait for all threads to complete and then delete them
    for (WidgetThreads::iterator iter = m_threads.begin();
         iter != m_threads.end(); ++iter)
    {
        (*iter)->stop();
        (*iter)->wait();
        delete *iter;
    }

    m_threads.clear();
}
