//==============================================================================
//
// Copyright (c) 2007 Ryan Winter. All rights reserved.
//
// This file may be used under the terms of the GNU General Public
// License version 2.0 appearing in the file LICENSE packaged with this file.
//
//==============================================================================

#include "fractal_widget.h"

#include <fractal/mandelbrot.h>
#include <config/processing_config.h>
#include <widget/rubber_band.h>

#include <QtGui>

#include <cmath>

namespace
{
    const double ZOOM_STEP = 2.0;
}

//------------------------------------------------------------------------------
// public:
//------------------------------------------------------------------------------
FractalWidget::FractalWidget(QWidget *parent)
    :
    QGraphicsView(parent),
    m_zoom(1.0),
    m_rotation(0.0),
    m_version(1),
    m_scene(),
    m_mode(Mode_Pan),
    m_rubber_band(new RubberBand(this)),
    m_bounding(),
    m_centre(),
    m_thread_pool()
{
    // connect the processing output
    connect(&m_thread_pool, SIGNAL(processingComplete(CellImage*)),
            this,           SLOT(addCell(CellImage*)));

    setScene(&m_scene);

    setResizeAnchor(QGraphicsView::AnchorViewCenter);
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

    m_scene.setBackgroundBrush(QColor(30, 30, 30));

    configure();
    panMode();

    buildScene();
}

//------------------------------------------------------------------------------
void
FractalWidget::configure()
{
    ProcessingConfig pc;
    m_thread_pool.setCpuNumber(pc.cpuNumber);
}

//------------------------------------------------------------------------------
// protected:
//------------------------------------------------------------------------------
void
FractalWidget::scrollContentsBy(int dx, int dy)
{
    QGraphicsView::scrollContentsBy(dx, dy);

    buildScene();
}

//------------------------------------------------------------------------------
void
FractalWidget::wheelEvent(QWheelEvent* event)
{
    // only handle vertical scrolling events
    if (event->orientation() != Qt::Vertical) return;

    // update the centre
    m_centre = mapToScene(viewport()->geometry().center());

    QPointF mouse_offset = mapToScene(event->pos()) - m_centre;

    if (event->delta() > 0)
    {
        // zoom in
        m_centre += (1.0 - 1.0 / ZOOM_STEP) * mouse_offset;
        zoom(ZOOM_STEP);
    }
    else if (event->delta() < 0)
    {
        // zoom out
        m_centre -= (ZOOM_STEP - 1.0) * mouse_offset;
        zoom(1.0 / ZOOM_STEP);
    }
}

//------------------------------------------------------------------------------
void
FractalWidget::showEvent(QShowEvent *event)
{
    buildScene();

    QGraphicsView::showEvent(event);
}

//------------------------------------------------------------------------------
void
FractalWidget::mouseMoveEvent(QMouseEvent* event)
{
    if (Mode_Zoom == m_mode)
    {
        m_rubber_band->update(viewport()->geometry(), event->pos());
    }

    QGraphicsView::mouseMoveEvent(event);
}

//------------------------------------------------------------------------------
void
FractalWidget::mousePressEvent(QMouseEvent* event)
{
    if (Mode_Zoom == m_mode)
    {
        m_rubber_band->setOrigin(event->pos());
    }

    QGraphicsView::mousePressEvent(event);
}

//------------------------------------------------------------------------------
void
FractalWidget::mouseReleaseEvent(QMouseEvent* event)
{
    if (Mode_Zoom == m_mode)
    {
        m_rubber_band->setDestination();
        zoom(m_rubber_band->geometry());
    }

    QGraphicsView::mouseReleaseEvent(event);
}

//------------------------------------------------------------------------------
// private slots:
//------------------------------------------------------------------------------
void
FractalWidget::panMode()
{
    setDragMode(QGraphicsView::ScrollHandDrag);
    m_mode = Mode_Pan;
}

//------------------------------------------------------------------------------
void
FractalWidget::zoomMode()
{
    setDragMode(QGraphicsView::NoDrag);
    m_mode = Mode_Zoom;
}

//------------------------------------------------------------------------------
void
FractalWidget::zoomIn()
{
    m_centre = mapToScene(viewport()->geometry().center());

    zoom(ZOOM_STEP);
}

//------------------------------------------------------------------------------
void
FractalWidget::zoomOut()
{
    m_centre = mapToScene(viewport()->geometry().center());

    zoom(1.0 / ZOOM_STEP);
}

//------------------------------------------------------------------------------
void
FractalWidget::addCell(CellImage* cell)
{
    CellDescriptor desc = cell->descriptor;

    if (desc.version == m_version)
    {
        QGraphicsPixmapItem* item
            = new QGraphicsPixmapItem(QPixmap::fromImage(cell->image));

        // put it in the correct place
        item->setPos(desc.x * desc.width, desc.y * desc.height);

        // add it to the scene
        m_scene.addItem(item);

        // force a repaint of the item (sometimes this doesnt happen)
        item->update();
    }

    delete cell;
}

//------------------------------------------------------------------------------
void
FractalWidget::buildScene()
{
    // setup the scroll limits
    Mandelbrot fractal;
    setSceneRect(fractal.limits(m_zoom));

    // grab the dimensions of the view
    const QPointF min = mapToScene(viewport()->geometry().topLeft());
    const QPointF max = mapToScene(viewport()->geometry().bottomRight());

    // calculate the viewable cells
    const int min_x = static_cast<int>(std::floor(min.x() / CellDescriptor::width));
    const int max_x = static_cast<int>(std::floor(max.x() / CellDescriptor::width));
    const int min_y = static_cast<int>(std::floor(min.y() / CellDescriptor::height));
    const int max_y = static_cast<int>(std::floor(max.y() / CellDescriptor::height));

    // generate the missing nodes
    for (int yy = min_y; yy <= max_y; ++yy)
    {
        for (int xx = min_x; xx <= max_x; ++xx)
        {
            // continue if we already contain this coordinate
            if (m_bounding.contains(xx, yy)) continue;

            // create the cell descriptor
            CellDescriptor cell;
            cell.x        = xx;
            cell.y        = yy;
            cell.zoom     = m_zoom;
            cell.rotation = m_rotation;
            cell.version  = m_version;

            // queue it
            m_thread_pool.queue(cell);
        }
    }

    // setup the new bounding
    m_bounding = QRect(QPoint(min_x, min_y), QPoint(max_x, max_y));
}

//------------------------------------------------------------------------------
// private:
//------------------------------------------------------------------------------
void
FractalWidget::zoom(float zoom)
{
    // dont zoom out past 1.0
    if (m_zoom * zoom < 1.0) return;

    ++m_version;

    // clear the old scene data
    clear();

    // calc the zoom and centre
    m_zoom   *= zoom;
    m_centre *= zoom;

    // rebuild the scene and re-centre
    buildScene();
    centerOn(m_centre);
}

//------------------------------------------------------------------------------
void
FractalWidget::zoom(QRect zoom_geometry)
{
    ++m_version;

    // clear the old scene data
    clear();

    // calculate the level of zoom
    float zoom = 1.0 * viewport()->width() / zoom_geometry.width();

    // calc the zoom and centre
    m_zoom   *= zoom;
    m_centre  = mapToScene(zoom_geometry.center());
    m_centre *= zoom;

    // rebuild the scene and re-centre
    buildScene();
    centerOn(m_centre);
}

//------------------------------------------------------------------------------
void
FractalWidget::clear()
{
    QList<QGraphicsItem*> items = m_scene.items();

    // clear all the items
    for (QList<QGraphicsItem*>::iterator iter = items.begin();
         iter != items.end(); ++iter)
    {
        delete *iter;
    }

    // reset to no bounding (force regenerate of whole fractal)
    m_bounding = QRect();
}
