//==============================================================================
//
// Copyright (c) 2007 Ryan Winter. All rights reserved.
//
// This file may be used under the terms of the GNU General Public
// License version 2.0 appearing in the file LICENSE packaged with this file.
//
//==============================================================================

#include "fractal_widget_thread.h"

#include <fractal/mandelbrot.h>
#include <fractal/cell_image.h>
#include <gradient/gradient.h>

//------------------------------------------------------------------------------
// public:
//------------------------------------------------------------------------------
FractalWidgetThread::FractalWidgetThread(CellDescriptorList& cells,
                                         QMutex&             lock,
                                         QSemaphore&         data_access)
    :
    QThread(),
    m_stop(false),
    m_cells(cells),
    m_lock(lock),
    m_data_access(data_access)
{
}

//------------------------------------------------------------------------------
FractalWidgetThread::~FractalWidgetThread()
{
}

//------------------------------------------------------------------------------
void
FractalWidgetThread::stop()
{
    m_stop = true;
}

//------------------------------------------------------------------------------
void
FractalWidgetThread::run()
{
    while (true)
    {
        // try and acquire a lock while polling completion
        while (!m_data_access.tryAcquire(1, 250))
        {
            if (m_stop) return;
        }

        // lock, then grab the item to process
        QMutexLocker ml(&m_lock);

            CellDescriptor cell = m_cells.back();
            m_cells.pop_back();

        ml.unlock();

        // generate the mandelbrot data
        std::vector<unsigned int> data(cell.width * cell.height);

        Mandelbrot mandel;
        mandel.generate(cell, data);

        // apply a gradient
        std::vector<Gradient::Point> points;
        points.push_back(Gradient::Point(0.0, QColor(0,   0, 0)));
        points.push_back(Gradient::Point(1.0, QColor(0, 255, 0)));
        Gradient g(points);

        std::vector<QRgb> gradient = g.swab(cell.iterations);

        CellImage *cell_image = new CellImage(cell);

        // fill out the image
        for (int xx = 0; xx < cell.width; ++xx)
        {
            for (int yy = 0; yy < cell.height; ++yy)
            {
                unsigned int p = data[yy * cell.width + xx];
                cell_image->image.setPixel(xx, yy, gradient[p]);
            }
        }

        // emit the result
        emit threadComplete(cell_image);
    }
}

